from daenerys.pipeline.html import ElementTreeProperty
from daenerys.pipeline.network import HTTPClientProperty, TextResponseProperty


class Dinergate(object):
    URL_TEMPLATE = None

    http_client = HTTPClientProperty()
    text_response = TextResponseProperty()
    etree = ElementTreeProperty()

    def __init__(self, request, http_client=None, **kwargs):
        self.request = request
        if http_client:
            self.http_client = http_client
        # assign arguments from URL pattern
        vars(self).update(kwargs)

    @property
    def url(self):
        if not self.URL_TEMPLATE:
            raise NotImplementedError
        return self.URL_TEMPLATE.format(self=self)

    def to_dict(self):
        return {k: getattr(self, k) for k in dir(self)
                if not k.startswith('_') and k not in ('text_response',)}



class Request(object):
    """The request object.

    :param url: the raw URL inputted from the dispatching app.
    :type url: :class:`urllib.parse.ParseResult`
    :param args: the query arguments decoded from query string of the URL.
    :type args: :class:`werkzeug.datastructures.MultiDict`
    """

    def __init__(self, url, args):
        self.url = url
        self.args = args

    def __repr__(self):
        return "Request(url={self.url}, args={self.args})".format(self=self)



class Site(object):
    def __init__(self, name):
        self.name = name
        self.actions = []

    def record_action(self, method_name, *args, **kwargs):
        self.actions.append((method_name, args, kwargs))

    def play_actions(self, target):
        for method_name, args, kwargs in self.actions:
            method = getattr(target, method_name)
            method(*args, **kwargs)

    def route(self, host, rule, **options):
        def decorator(func):
            endpoint = "{func.__module__}:{func.__name__}".format(func=func)
            self.record_action("add_url_rule", host, rule, endpoint, **options)
            return func
        return decorator



import lxml.html

from daenerys.pipeline.base import PipelineProperty


class ElementTreeProperty(PipelineProperty):
    def prepare(self):
        self.options.setdefault("encoding", None)

    def provide_value(self, obj):
        text_response = obj.text_response
        if self.options["encoding"]:
            text_response = text_response.encode(self.options["encoding"])
        return lxml.html.fromstring(text_response)


class XPathTextProperty(PipelineProperty):
    required_attrs = {"xpath"}
    def prepare(self):
        self.options.setdefault("strip_spaces", False)
        self.options.setdefault("pick_mode", "join")
        self.options.setdefault("joiner", " ")
        self.options.setdefault("namespaces", None)

    def choice_pick_impl(self):
        pick_mode = self.options["pick_mode"]
        impl = {
            "join": self.pick_joining,
            "first": self.pick_first,
            "keep": self.keep_value,
        }.get(pick_mode)

        if not impl:
            raise ValueError("%r is not valid pick mode" % pick_mode)
        return impl

    def pick_joining(self, value):
        joiner = self.options["joiner"]
        return joiner.join(value)

    def pick_first(self, value):
        return value[0] if value else ""

    def keep_value(self, value):
        return value

    def provide_value(self, obj):
        value = obj.etree.xpath(
            self.xpath, namespaces=self.options.get("namespaces"))
        pick_value = self.choice_pick_impl()

        if self.options["strip_spaces"]:
            value = [v.strip() for v in value if v.strip()]

        return pick_value(value)



from .base import PipelineProperty
from .html import ElementTreeProperty, XPathTextProperty
from .network import (HTTPClientProperty, TextResponseProperty,
                      JSONResponseProperty)


__all__ = ["PipelineProperty", "ElementTreeProperty", "XPathTextProperty",
           "HTTPClientProperty", "TextResponseProperty",
           "JSONResponseProperty"]



import urllib
from urlparse import urlparse, ParseResult

from werkzeug.utils import find_modules, import_string
from werkzeug.urls import url_decode, url_encode
from werkzeug.routing import Map, Rule, NotFound, RequestRedirect

from .request import Request
from .exceptions import NotSupported


class Daenerys(object):
    def __init__(self, ignore_sites=set()):
        self.url_map = Map(strict_slashes=False, host_matching=True,
                           redirect_defaults=False)
        self.ignore_sites = ignore_sites

    def add_url_rule(self, host, rule_string, endpoint, **options):
        rule = Rule(rule_string, host=host, endpoint=endpoint, **options)
        self.url_map.add(rule)

    def parse_url(self, url_string):
        url = urlparse(url_string)
        url = self.validate_url(url)
        url_adapter = self.url_map.bind(server_name=url.hostname,
                                        url_scheme=url.scheme,
                                        path_info=url.path)
        query_args = url_decode(url.query)
        return url, url_adapter, query_args

    def validate_url(self, url):
        url_path = urllib.quote(url.path, safe=b"/%")
        url_query = urllib.quote(url.query, safe=b"?=&")

        url = ParseResult(url.scheme, url.netloc, url_path,
                          url.params, url_query, url.fragment)

        has_hostname = url.hostname is not None and len(url.hostname) > 0
        has_http_scheme = url.scheme in ("http", "https")
        has_path = not len(url.path) or url.path.startswith("/")

        if not (has_hostname and has_http_scheme and has_path):
            raise NotSupported("invalid url: %s" % repr(url))

        return url

    def dispatch_url(self, url_string):
        url, url_adapter, query_args = self.parse_url(url_string)

        try:
            endpoint, kwargs = url_adapter.match()
        except NotFound:
            raise NotSupported(url_string)
        except RequestRedirect as e:
            new_url = "{0.new_url}?{1}".format(e, url_encode(query_args))
            return self.dispatch_url(new_url)

        try:
            handler = import_string(endpoint)
            request = Request(url=url, args=query_args)
            return handler(request, **kwargs)
        except RequestRedirect as e:
            return self.dispatch_url(e.new_url)

    def mount_site(self, site):
        if isinstance(site, basestring):
            site = import_string(site)
        site.play_actions(target=self)

    def mount_sites(self, root):
        for name in find_modules(root, recursive=True):
            mod = import_string(name)
            site = name.split('.')[-1]
            if hasattr(mod, 'site') and site not in self.ignore_sites:
                mod.site.play_actions(target=self)




# coding=utf-8
from os.path import join, dirname
from setuptools import setup, find_packages

VERSION = (0, 1)
__version__ = VERSION
__versionstr__ = '.'.join(map(str, VERSION))

with open(join(dirname(__file__), 'README.md')) as f:
    long_description = f.read().strip()

setup(
    name = 'daenerys',
    description = 'Scraping and Web Crawling Framework',
    url = 'https://code.pycourses.com/courses/daenerys',
    long_description = long_description,
    version = __versionstr__,
    author = 'Dong Weiming',
    author_email = '61966225@qq.com',
    packages=find_packages(),
    classifiers = [
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
    ]
)



from requests import Session

from daenerys.pipeline.base import PipelineProperty
from daenerys.exceptions import NotSupported


class HTTPClientProperty(PipelineProperty):
    def prepare(self):
        self.options.setdefault("session_class", Session)

    def provide_value(self, obj):
        session_class = self.options["session_class"]
        session = session_class()
        return session


class ResponseProperty(PipelineProperty):
    def prepare(self):
        self.options.setdefault("method", "GET")
        self.options.setdefault("data", {})

    def provide_value(self, obj):
        if "content_method" not in self.options:
            raise KeyError("You need create a subclass which inheritance "
                           "ResponseProperty, and assign `content_method` "
                           "into self.attr_names")
        response = obj.http_client.request(
            url=obj.url, method=self.options.get('method'),
            **self.options['data'])
        response.raise_for_status()
        content = getattr(response, self.options.get('content_method'))
        if callable(content):
            content = content()
        return content


class TextResponseProperty(ResponseProperty):
    def prepare(self):
        super(TextResponseProperty, self).prepare()
        self.options.setdefault("content_method", "content")


class JSONResponseProperty(ResponseProperty):
    def prepare(self):
        super(JSONResponseProperty, self).prepare()
        self.options.setdefault("content_method", "json")



import abc


class _Missing(object):
    def __repr__(self):
        return 'no value'
    def __reduce__(self):
        return '_missing'


_missing = _Missing()


class PipelineProperty:
    __metaclass__ = abc.ABCMeta
    __counter = 0
    required_attrs = set()

    def __init__(self, **options):
        cls = self.__class__
        self.options = options
        prefix = cls.__name__
        index = cls.__counter
        self.name = '_{}#{}'.format(prefix, index)
        cls.__counter += 1

        assigned_attrs = set()
        for name, value in options.items():
            assigned_attrs.add(name)

            # required attrs
            if name in self.required_attrs:
                setattr(self, name, value)
        missing_attrs = self.required_attrs - assigned_attrs
        if missing_attrs:
            raise TypeError("missing %r" % ", ".join(missing_attrs))

        self.prepare()

    def __set__(self, obj, value):
        obj.__dict__[self.name] = value

    def __get__(self, obj, type=None):
        if obj is None:
            return self
        value = obj.__dict__.get(self.name, _missing)
        if value is _missing:
            value = self.provide_value(obj)
            obj.__dict__[self.name] = value
        return value

    def prepare(self):
        """This method will be called after instance ininialized. The
        subclasses could override the implementation."""

    @abc.abstractmethod
    def provide_value(self, obj):
        pass
